import factory
from factory import fuzzy

from django.core.urlresolvers import reverse
from django.utils import timezone

from nodeconductor.structure.tests import factories as structure_factories

from waldur_paypal import models


class PaypalPaymentFactory(factory.DjangoModelFactory):
    class Meta(object):
        model = models.Payment

    amount = 10
    customer = factory.SubFactory(structure_factories.CustomerFactory)
    backend_id = factory.Sequence(lambda n: 'PAYMENT-ABC-%s' % n)
    token = factory.Sequence(lambda n: 'TOKEN-%s' % n)

    @classmethod
    def get_url(self, payment=None, action=None):
        if payment is None:
            payment = PaypalPaymentFactory()
        url = 'http://testserver' + reverse('paypal-payment-detail', kwargs={'uuid': payment.uuid})
        return url if action is None else url + action + '/'

    @classmethod
    def get_list_url(cls):
        return 'http://testserver' + reverse('paypal-payment-list')


class InvoiceFactory(factory.DjangoModelFactory):
    class Meta(object):
        model = models.Invoice

    customer = factory.SubFactory(structure_factories.CustomerFactory)
    backend_id = factory.Sequence(lambda n: 'INV2-ETBW-Q5NB-VWLT-9RH%s' % n)
    state = models.Invoice.States.DRAFT
    invoice_date = factory.fuzzy.FuzzyDate(start_date=timezone.now().date())
    month = factory.fuzzy.FuzzyInteger(1, 12)
    year = factory.fuzzy.FuzzyInteger(1970, 2017)
    end_date = factory.fuzzy.FuzzyDate(start_date=timezone.now().date())
    tax_percent = factory.fuzzy.FuzzyInteger(1, 10)
    issuer_details = factory.Dict({
        'email': factory.Sequence(lambda n: 'email-%s@domain.com' % n),
        'address': factory.Sequence(lambda n: 'address-%s' % n),
        'city': factory.Sequence(lambda n: 'city-%s' % n),
        'postal_code': 1101,
        'country_code': 'EE',
        'phone': {
            'country_code': '372',
            'national_number': '5555555'
        }
    })


class InvoiceItemFactory(factory.DjangoModelFactory):
    class Meta(object):
        model = models.InvoiceItem

    invoice = factory.SubFactory(InvoiceFactory)
    price = factory.fuzzy.FuzzyInteger(10, 100)
    tax = factory.fuzzy.FuzzyInteger(1, 10)
    unit_price = factory.fuzzy.FuzzyInteger(5, 30)
    quantity = factory.fuzzy.FuzzyInteger(24, 2400, step=24)
    unit_of_measure = models.InvoiceItem.UnitsOfMeasure.HOURS
    name = factory.Sequence(lambda n: 'Invoice-item-%s' % n)
    start = factory.fuzzy.FuzzyDateTime(start_dt=timezone.now())



from django.test import TestCase
import mock

from waldur_paypal.backend import PaypalBackend, PayPalError

from .. import factories, fixtures


class BaseBackendTest(TestCase):

    def setUp(self):
        self.fixture = fixtures.PayPalFixture()
        self.backend = PaypalBackend()
        self.invoice = self.fixture.invoice


class CreateInvoiceTest(BaseBackendTest):

    def setUp(self):
        super(CreateInvoiceTest, self).setUp()
        self.invoice.backend_id = ''
        self.invoice.save()

    @mock.patch('waldur_paypal.backend.paypal.Invoice')
    def test_invoice_is_created(self, invoice_mock):
        self.invoice.items.add(factories.InvoiceItemFactory())
        backend_invoice = mock.Mock()
        backend_invoice.id = 'INV2-ETBW-Q5NB-VWLT-9RH1'
        backend_invoice.status = 'DRAFT'
        backend_invoice.number = '00001'
        invoice_mock.return_value = backend_invoice

        result = self.backend.create_invoice(self.invoice)

        self.assertEqual(result.backend_id, backend_invoice.id)
        self.assertEqual(result.state, backend_invoice.status)
        backend_invoice.create.assert_called_once()

    def test_invoice_is_not_created_if_phone_is_not_provided(self):
        self.invoice.items.add(factories.InvoiceItemFactory())
        del self.invoice.issuer_details['phone']

        self.assertRaises(PayPalError, self.backend.create_invoice, self.invoice)

    def test_invoice_is_not_created_if_phone_country_code_is_not_provided(self):
        self.invoice.items.add(factories.InvoiceItemFactory())
        del self.invoice.issuer_details['phone']['country_code']

        self.assertRaises(PayPalError, self.backend.create_invoice, self.invoice)

    def test_invoice_is_not_created_if_phone_national_number_is_not_provided(self):
        self.invoice.items.add(factories.InvoiceItemFactory())
        del self.invoice.issuer_details['phone']['national_number']

        self.assertRaises(PayPalError, self.backend.create_invoice, self.invoice)


class DownloadInvoicePDFTest(BaseBackendTest):

    def setUp(self):
        super(DownloadInvoicePDFTest, self).setUp()

    def test_pdf_is_not_downloaded_if_backend_id_is_missing(self):
        self.invoice.backend_id = ''
        self.invoice.save()

        self.assertRaises(PayPalError, self.backend.download_invoice_pdf, self.invoice)

    @mock.patch('waldur_paypal.backend.urllib2.urlopen')
    def test_pdf_is_downloaded(self, urlopen_mock):
        response = mock.Mock()
        response.read.return_value = 'PDF file content'
        urlopen_mock.return_value = response
        self.assertFalse(self.invoice.pdf)

        self.backend.download_invoice_pdf(self.invoice)

        self.assertTrue(self.invoice.pdf)
        urlopen_mock.assert_called_once()


class SendInvoiceTest(BaseBackendTest):

    @mock.patch('waldur_paypal.backend.paypal.Invoice')
    def test_draft_invoice_is_sent(self, invoice_mock):
        self.invoice.items.add(factories.InvoiceItemFactory())
        backend_invoice = mock.Mock(name='backend_invoice')
        invoice_mock.find.return_value = backend_invoice

        self.backend.send_invoice(self.invoice)

        backend_invoice.send.assert_called_once()

    def test_invoice_cannot_be_sent_if_it_is_not_in_draft_state(self):
        self.invoice.state = self.invoice.States.SENT
        self.invoice.save()

        self.assertRaises(PayPalError, self.backend.send_invoice, self.invoice)



import decimal
import mock

from rest_framework import test, status

from nodeconductor.structure.tests import fixtures as structure_fixtures, factories as structure_factories
from waldur_paypal.backend import PaypalPayment, PayPalError
from waldur_paypal.helpers import override_paypal_settings
from waldur_paypal.models import Payment

from . import factories


@override_paypal_settings(ENABLED=True)
class BasePaymentTest(test.APITransactionTestCase):

    def setUp(self):
        self.fixture = structure_fixtures.CustomerFixture()
        self.customer = self.fixture.customer
        self.other = structure_factories.UserFactory()

        self.valid_request = {
            'amount': decimal.Decimal('9.99'),
            'customer': structure_factories.CustomerFactory.get_url(self.customer),
            'return_url': 'http://example.com/return/',
            'cancel_url': 'http://example.com/cancel/'
        }

        self.valid_response = {
            'approval_url': 'https://www.paypal.com/webscr?cmd=_express-checkout&token=EC-60U79048BN7719609',
            'payer_id': '7E7MGXCWTTKK2',
            'token': 'EC-60U79048BN7719609'
        }


class PaymentCreateTest(BasePaymentTest):

    def create_payment(self, user, fail=False):
        with mock.patch('waldur_paypal.backend.PaypalBackend') as backend:
            if fail:
                backend().make_payment.side_effect = PayPalError()
            else:
                backend().make_payment.return_value = PaypalPayment(
                        payment_id='PAY-6RV70583SB702805EKEYSZ6Y',
                        approval_url=self.valid_response['approval_url'],
                        token=self.valid_response['token'])

            self.client.force_authenticate(user)
            return self.client.post(factories.PaypalPaymentFactory.get_list_url(), data=self.valid_request)

    def test_staff_can_create_payment_for_any_customer(self):
        response = self.create_payment(self.fixture.staff)
        self.assertEqual(response.status_code, status.HTTP_201_CREATED, response.data)
        self.assertEqual(self.valid_response['approval_url'], response.data['approval_url'])

    def test_user_can_create_payment_for_owned_customer(self):
        response = self.create_payment(self.fixture.owner)
        self.assertEqual(response.status_code, status.HTTP_201_CREATED, response.data)

    def test_user_can_not_create_payment_for_other_customer(self):
        response = self.create_payment(self.other)
        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)

    def test_when_backend_fails_error_returned(self):
        response = self.create_payment(self.fixture.owner, fail=True)
        self.assertEqual(response.status_code, status.HTTP_500_INTERNAL_SERVER_ERROR)

    @override_paypal_settings(ENABLED=False)
    def test_failed_dependency_raised_if_extension_is_disabled(self):
        response = self.create_payment(self.fixture.owner)
        self.assertEqual(response.status_code, status.HTTP_424_FAILED_DEPENDENCY)


class PaymentApprovalTest(BasePaymentTest):

    def approve_payment(self, user, amount=None, fail=False):
        payment = factories.PaypalPaymentFactory(
            customer=self.customer,
            state=Payment.States.CREATED,
            amount=amount or 100.0)

        with mock.patch('waldur_paypal.backend.PaypalBackend') as backend:
            if fail:
                backend().approve_payment.side_effect = PayPalError()

            self.client.force_authenticate(user)
            return self.client.post(factories.PaypalPaymentFactory.get_list_url() + 'approve/', data={
                'payment_id': payment.backend_id,
                'payer_id': self.valid_response['payer_id'],
                'token': payment.token
            })

    def test_staff_can_approve_any_payment(self):
        response = self.approve_payment(self.fixture.staff)
        self.assertEqual(response.status_code, status.HTTP_200_OK, response.data)

    def test_user_can_approve_payment_for_owned_customer(self):
        response = self.approve_payment(self.fixture.owner)
        self.assertEqual(response.status_code, status.HTTP_200_OK, response.data)

    def test_user_can_not_approve_payment_for_other_customer(self):
        response = self.approve_payment(self.other)
        self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)


class PaymentCancellationTest(BasePaymentTest):

    def cancel_payment(self, user):
        self.client.force_authenticate(user)
        payment = factories.PaypalPaymentFactory(customer=self.customer, state=Payment.States.CREATED)
        return self.client.post(factories.PaypalPaymentFactory.get_list_url() + 'cancel/', data={
            'token': payment.token
        })

    def test_staff_can_cancel_any_payment(self):
        response = self.cancel_payment(self.fixture.staff)
        self.assertEqual(response.status_code, status.HTTP_200_OK, response.data)

    def test_user_can_cancel_payment_for_owned_customer(self):
        response = self.cancel_payment(self.fixture.owner)
        self.assertEqual(response.status_code, status.HTTP_200_OK)

    def test_user_can_not_cancel_payment_for_other_customer(self):
        response = self.cancel_payment(self.other)
        self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)

    @override_paypal_settings(ENABLED=False)
    def test_failed_dependency_raised_if_extension_is_disabled(self):
        response = self.cancel_payment(self.fixture.owner)
        self.assertEqual(response.status_code, status.HTTP_424_FAILED_DEPENDENCY)



from django.utils.functional import cached_property

from nodeconductor.structure.tests import fixtures as structure_fixtures

from . import factories


class PayPalFixture(structure_fixtures.CustomerFixture):

    @cached_property
    def payment(self):
        return factories.PaypalPaymentFactory(customer=self.customer)

    @cached_property
    def invoice(self):
        return factories.InvoiceFactory(customer=self.customer)



import datetime
import mock

from django.test import TestCase
from django.utils import timezone

from waldur_paypal import handlers, models

from .. import fixtures


class CreateInvoiceTest(TestCase):

    def setUp(self):
        self.fixture = fixtures.PayPalFixture()
        self.issuer_details = {
            'email': "example@domain.com",
            'first_name': "John",
            'last_name': "White",
            'business_name': "Corporation Professionals, LLC",
            'phone': {
                'country_code': '001',
                'national_number': '5032141716',
            },
            'address': {
                "line1": "1234 Main St.",
                "city": "Portland",
                "state": "OR",
                "postal_code": "97217",
                "country_code": "US"
            }
        }

    def _get_valid_invoice(self):
        invoice = mock.Mock()
        invoice.customer = self.fixture.customer
        invoice.year = 2017
        invoice.month = 1
        invoice.tax_percent = 15

        # when invoice is moved to created state
        invoice.invoice_date = timezone.now().date()
        invoice.due_date = invoice.invoice_date + datetime.timedelta(days=30)
        return invoice

    def _get_payment_details(self):
        payment_details = mock.Mock()
        payment_details.company = 'Company'
        payment_details.address = 'Address #1'
        payment_details.country = 'Wonderland'
        payment_details.email = 'test@example.com'
        payment_details.postal = '10110'
        payment_details.phone = '+372-555-55-55'
        payment_details.bank = 'Harry Potter Bank #2.5'
        return payment_details

    def _generate_invoice_items(self, count=1):
        items = []

        for i in range(0, count):
            item = mock.Mock()
            item.unit = 'day'
            item.unit_price = 10
            item.usage_days = 15
            item.price = item.unit_price * item.usage_days
            item.tax = 15
            item.start = timezone.now()
            item.end = item.start + timezone.timedelta(days=item.usage_days)
            item.name = 'invoice item #%s' % i
            items.append(item)

        return items

    def test_invoice_is_created(self):
        invoice = self._get_valid_invoice()
        invoice.items = self._generate_invoice_items(2)
        self.assertEqual(models.Invoice.objects.count(), 0)

        handlers.create_invoice(sender=None, invoice=invoice, issuer_details=self.issuer_details)

        self.assertEqual(models.Invoice.objects.count(), 1)
        new_invoice = models.Invoice.objects.get(invoice_date=invoice.invoice_date, customer=invoice.customer)
        self.assertEqual(new_invoice.year, invoice.year)
        self.assertEqual(new_invoice.month, invoice.month)
        self.assertEqual(new_invoice.items.count(), 2)

        for original_item in invoice.items:
            created_item = [item for item in new_invoice.items.iterator() if item.name == original_item.name][0]
            self.assertEqual(created_item.unit_price, original_item.unit_price)
            self.assertEqual(created_item.price, original_item.price)
            self.assertEqual(created_item.start, original_item.start)
            self.assertEqual(created_item.end, original_item.end)
            self.assertEqual(created_item.tax, original_item.tax)
            self.assertEqual(created_item.unit_of_measure, models.InvoiceItem.UnitsOfMeasure.HOURS)
            self.assertEqual(created_item.quantity, original_item.usage_days * 24)




import logging

from nodeconductor.logging import models as logging_models

from .log import event_logger
from . import models, helpers


logger = logging.getLogger(__name__)


def log_invoice_save(sender, instance, created=False, **kwargs):
    if created:
        event_logger.paypal_invoice.info(
            '{invoice_invoice_date}-{invoice_end_date}. Invoice for customer {customer_name} has been created.',
            event_type='invoice_creation_succeeded',
            event_context={
                'invoice': instance,
            })
    else:
        event_logger.paypal_invoice.info(
            '{invoice_invoice_date}-{invoice_end_date}. Invoice for customer {customer_name} has been updated.',
            event_type='invoice_update_succeeded',
            event_context={
                'invoice': instance,
            })


def log_invoice_delete(sender, instance, **kwargs):
    event_logger.paypal_invoice.info(
        '{invoice_invoice_date}-{invoice_end_date}. Invoice for customer {customer_name} has been deleted.',
        event_type='invoice_deletion_succeeded',
        event_context={
            'invoice': instance,
        })


def add_email_hooks_to_user(sender, instance, created, **kwargs):
    if not created:
        return
    event_types = ['invoice_creation_succeeded', 'payment_creation_succeeded',
                   'payment_approval_succeeded', 'payment_cancel_succeeded']
    user = instance
    if not user.email:
        logger.warn('Cannot add default email hooks to user %s (PK: %s). He does not have email.', user, user.pk)

    logging_models.EmailHook.objects.create(
        user=user,
        event_types=event_types,
        email=user.email,
    )


def create_invoice(sender, invoice, issuer_details, **kwargs):
    """
    Creates an invoice when customer is "billed".
    :param sender: Invoice model
    :param invoice: Invoice instance
    :param issuer_details: details about issuer
    """
    paypal_invoice = models.Invoice(
        customer=invoice.customer,
        year=invoice.year,
        month=invoice.month,
        invoice_date=invoice.invoice_date,
        end_date=invoice.due_date,
        tax_percent=invoice.tax_percent,
        issuer_details=issuer_details)

    if hasattr(invoice.customer, 'payment_details'):
        payment_details = invoice.customer.payment_details
        paypal_invoice.payment_details = {
            'company': payment_details.company,
            'address': payment_details.address,
            'country': payment_details.country,
            'email': payment_details.email,
            'postal': payment_details.postal,
            'phone': payment_details.phone,
            'bank': payment_details.bank,
        }

    paypal_invoice.save()

    for item in invoice.items:
        models.InvoiceItem.objects.create(
            invoice=paypal_invoice,
            price=item.price,
            tax=item.tax,
            quantity=helpers.get_invoice_item_quantity(item),
            unit_price=item.unit_price,
            unit_of_measure=helpers.convert_unit_of_measure(item.unit),
            name=item.name,
            start=item.start,
            end=item.end)



import copy

from django.conf import settings
from django.test import override_settings

from . import models


def override_paypal_settings(**kwargs):
    plugin_settings = copy.deepcopy(settings.WALDUR_PAYPAL)
    plugin_settings.update(kwargs)
    return override_settings(WALDUR_PAYPAL=plugin_settings)


def convert_unit_of_measure(unit):
    if unit.lower() == 'quantity':
        return models.InvoiceItem.UnitsOfMeasure.QUANTITY
    else:
        return models.InvoiceItem.UnitsOfMeasure.HOURS


def get_invoice_item_quantity(item):
    unit_of_measure = convert_unit_of_measure(item.unit)
    if unit_of_measure == models.InvoiceItem.UnitsOfMeasure.HOURS:
        return item.usage_days * 24
    else:
        return item.quantity
